import uuid
from enum import IntEnum
from pathlib import Path

from pydantic import BaseModel


class Category(IntEnum):
    pwn = 1
    web = 2
    rev = 3
    crypto = 4
    steg = 5
    misc = 6
    osint = 7
    forensics = 8
    hardware = 9
    mobile = 10
    game = 11
    blockchain = 12
    other = 13


class Solved(IntEnum):
    unsolved = 0
    solved = 1


class Challenge(BaseModel):
    id: uuid.UUID | int = uuid.uuid4()
    name: str
    flag: str | None = None
    points: int | None = None
    solved: Solved = Solved.unsolved
    category: Category = Category.pwn
    ctf_id: uuid.UUID
    path: Path
    tags: list[str] | None = None
    extras: dict | None = None



from pathlib import Path

import typer

from pwnv.cli.utils import (
    add_ctf,
    add_remote_ctf,
    config_exists,
    ctfs_exists,
    error,
    get_ctfs,
    get_ctfs_path,
    get_current_ctf,
    get_running_ctfs,
    get_stopped_ctfs,
    is_duplicate,
    prompt_confirm,
    prompt_ctf_selection,
    prompt_text,
    remove_ctf,
    sanitize,
    show_ctf,
    success,
    update_ctf,
    warn,
)
from pwnv.models import CTF
from pwnv.models.ctf import Status

app = typer.Typer(no_args_is_help=True)


@app.command()
@config_exists()
def add(name: str):
    path: Path = (get_ctfs_path() / sanitize(name)).resolve()
    if is_duplicate(path=path, model_list=get_ctfs()):
        error(f"CTF [cyan]{name}[/] already exists.")

        return

    if prompt_confirm(
        "Do you want to add a remote CTF? (y/n)",
        default=False,
    ):
        add_remote_ctf(CTF(name=name, path=path, url=prompt_text("Enter the URL:")))
    else:
        add_ctf(CTF(name=name, path=path))
    success(f"CTF [cyan]{name}[/] added.")


@app.command()
@config_exists()
@ctfs_exists()
def remove():
    chosen_ctf = prompt_ctf_selection(get_ctfs(), "Select a CTF to remove:")
    if not prompt_confirm(
        f"Remove CTF '{chosen_ctf.name}' and all its challenges?",
        default=False,
    ):
        return
    remove_ctf(chosen_ctf)
    success(f"CTF [cyan]{chosen_ctf.name}[/] removed")


@app.command()
@config_exists()
@ctfs_exists()
def info():
    while True:
        ctfs: list[CTF] = get_ctfs()
        show_ctf(prompt_ctf_selection(ctfs, "Select a CTF to show info:"))
        if not prompt_confirm("Show another CTF?", default=False):
            break


@app.command()
@config_exists()
@ctfs_exists()
def stop():
    running: list[CTF] = get_running_ctfs()
    if not running:
        warn("No running CTFs found.")

        return
    current = get_current_ctf()
    if current in running:
        chosen_ctf = current
    else:
        chosen_ctf = prompt_ctf_selection(running, "Select a CTF to stop:")
    chosen_ctf.running = Status.stopped
    update_ctf(chosen_ctf)
    success(f"CTF [cyan]{chosen_ctf.name}[/] stopped.")


@app.command()
@config_exists()
@ctfs_exists()
def start():
    stopped: list[CTF] = get_stopped_ctfs()
    if not stopped:
        warn("No stopped CTFs found.")

        return
    current = get_current_ctf()
    if current in stopped:
        chosen_ctf = current
    else:
        chosen_ctf = prompt_ctf_selection(stopped, "Select a CTF to start:")
    chosen_ctf.running = Status.running
    update_ctf(chosen_ctf)
    success(f"CTF [cyan]{chosen_ctf.name}[/] started.")



import asyncio

import typer

from pwnv.cli.utils import (
    add_tags,
    challenges_exists,
    config_exists,
    get_ctf_by_challenge,
    get_current_challenge,
    get_unsolved_challenges,
    prompt_challenge_selection,
    prompt_text,
    remote_solve,
    success,
    update_challenge,
    warn,
)
from pwnv.models.challenge import Solved

app = typer.Typer(no_args_is_help=True)


@app.command()
@config_exists()
@challenges_exists()
def solve(flag: str = "") -> None:
    unsolved = get_unsolved_challenges()
    if not unsolved:
        warn("No unsolved challenges found.")

        return

    challenge = (
        get_current_challenge() if get_current_challenge() in unsolved else None
    ) or prompt_challenge_selection(unsolved, "Select a challenge:")

    challenge.solved = Solved.solved
    if not flag:
        flag = prompt_text("Enter the flag:")

    if flag:
        challenge.flag = flag

    ctf = get_ctf_by_challenge(challenge)
    if (ctf.path / ".env").exists():
        if not asyncio.run(remote_solve(challenge=challenge, ctf=ctf, flag=flag)):
            return
    raw = prompt_text("Enter tags (comma-separated):")
    if raw:
        tags = {t.strip().lower() for t in raw.split(",") if t.strip()}
        add_tags(tags)
        challenge.tags = sorted(tags)

    update_challenge(challenge)
    success(f"[cyan]{challenge.name}[/] marked as solved.")



from pwnv.models import Challenge
from pwnv.models.challenge import Category
from pwnv.plugins import PwnPlugin
from pwnv.plugins.plugin import ChallengePlugin

PLUGIN_REGISTRY: dict[Category, ChallengePlugin] = {
    Category.pwn: PwnPlugin(),
    # add other plugins here
}


class Core:
    def __init__(self, challenge: Challenge):
        plugin = PLUGIN_REGISTRY.get(challenge.category)
        if plugin:
            plugin.create_template(challenge)
            plugin.logic(challenge)



from typing import Annotated, List

import typer

from pwnv.cli.utils import (
    add_challenge,
    challenges_exists,
    challenges_for_ctf,
    config_exists,
    ctfs_exists,
    error,
    get_challenges,
    get_ctfs,
    get_current_challenge,
    get_current_ctf,
    get_running_ctfs,
    get_solved_challenges,
    is_duplicate,
    prompt_category_selection,
    prompt_challenge_selection,
    prompt_confirm,
    prompt_ctf_selection,
    prompt_tags_selection,
    remove_challenge,
    sanitize,
    show_challenge,
    success,
    warn,
)
from pwnv.models import CTF, Challenge

app = typer.Typer(no_args_is_help=True)


@app.command()
@config_exists()
@ctfs_exists()
def add(name: str) -> None:
    chosen_ctf: CTF | None = get_current_ctf() or (
        prompt_ctf_selection(get_running_ctfs(), "Select a running CTF:")
        if get_running_ctfs()
        else None
    )
    if not chosen_ctf:
        warn("No running CTFs found.")

        return

    category = prompt_category_selection()
    ch_path = chosen_ctf.path / category.name / sanitize(name)

    if ch_path.exists() or is_duplicate(
        path=ch_path, model_list=challenges_for_ctf(chosen_ctf)
    ):
        error(
            f"[cyan]{name}[/] already exists in "
            f"[cyan]{chosen_ctf.name}/{category.name}/[/]."
        )
        return

    challenge = Challenge(
        ctf_id=chosen_ctf.id, name=name, path=ch_path, category=category
    )
    add_challenge(challenge)
    success(f"[cyan]{challenge.name}[/] added")


@app.command()
@config_exists()
@challenges_exists()
def remove() -> None:
    challenges: List[Challenge] = (
        challenges_for_ctf(get_current_ctf()) if get_current_ctf() else get_challenges()
    )
    challenge = prompt_challenge_selection(challenges, "Select a challenge to remove:")

    if challenge.path.exists() and any(challenge.path.iterdir()):
        if not prompt_confirm("Directory not empty. Remove anyway?", default=False):
            return

    remove_challenge(challenge)
    success(f"[cyan]{challenge.name}[/] removed")


@app.command()
@config_exists()
@challenges_exists()
def info(
    all: Annotated[bool, typer.Option(help="Show challenges from all CTFs")] = False,
) -> None:
    current = get_current_challenge()
    if current:
        show_challenge(current)
        return
    challenges = (
        get_challenges()
        if all
        else (
            challenges_for_ctf(
                get_current_ctf()
                if get_current_ctf()
                else prompt_ctf_selection(get_ctfs(), "Select a CTF:")
            )
        )
    )

    if not challenges:
        warn("No challenges found.")
        return

    while True:
        show_challenge(prompt_challenge_selection(challenges, "Select a challenge:"))
        if not prompt_confirm("Show another?", default=False):
            break


@app.command(name="filter")
@config_exists()
@challenges_exists()
def filter_() -> None:
    solved = get_solved_challenges()
    if not solved:
        warn("No solved challenges found.")
        return

    while True:
        tags = prompt_tags_selection("Select tags to filter by:")
        subset = [ch for ch in solved if ch.tags and any(t in ch.tags for t in tags)]
        if not subset:
            warn("No challenges match your tags.")

        else:
            show_challenge(prompt_challenge_selection(subset, "Select a challenge:"))
        if not prompt_confirm("Filter again?", default=False):
            break



def main():
    import typer

    from pwnv.cli import challenge_app, ctf_app, init_app, reset_app, solve_app

    app = typer.Typer(no_args_is_help=True)
    app.add_typer(challenge_app, name="challenge")
    app.add_typer(ctf_app, name="ctf")
    app.add_typer(init_app)
    app.add_typer(reset_app)
    app.add_typer(solve_app)
    app()



import os
import shutil
import subprocess
from pathlib import Path
from typing import Annotated, Optional

import typer

from pwnv.cli.utils import (
    command,
    error,
    get_config_path,
    info,
    prompt_confirm,
    save_config,
    warn,
)
from pwnv.models import Init

app = typer.Typer(no_args_is_help=True)

_PKGS = [
    "pwntools",
    "ropgadget",
    "angr",
    "spwn",
    "pycryptodome",
    "z3",
    "requests",
    "libdebug",
]


@app.command()
def init(
    ctfs_folder: Annotated[
        Optional[Path], typer.Option(help="Directory that will store all CTFs")
    ] = Path.cwd() / "CTF",
) -> None:
    if not shutil.which("uv"):
        error(f"{command('uv')} binary not found in PATH. Install it first.")
        return

    cfg_path = get_config_path()
    if cfg_path.exists():
        error("Config file already exists - aborting.")
        return

    ctfs_folder = ctfs_folder.resolve()
    env_path = ctfs_folder / ".pwnvenv"

    if ctfs_folder.exists() and ctfs_folder.iterdir():
        if not prompt_confirm(
            f"Directory {ctfs_folder} already exists. Continue?", default=False
        ):
            return
    else:
        if not prompt_confirm(
            f"Create new CTF directory at {ctfs_folder}?", default=True
        ):
            return

    ctfs_folder.mkdir(parents=True, exist_ok=True)
    init_model = Init(ctfs_path=ctfs_folder, challenge_tags=[], ctfs=[], challenges=[])
    save_config(init_model.model_dump())

    if (
        not subprocess.run(
            ["uv", "venv", str(env_path)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode
        == 0
    ):
        error("Failed to initialise uv environment.")

        return
    os.chdir(ctfs_folder)
    if not subprocess.run(["uv", "pip", "install", *_PKGS]).returncode == 0:
        warn("Failed to add default packages.")

        return
    info(f"Activate with {command(f'source {env_path}/bin/activate')}. Happy hacking!")



from pwnv.cli.utils.config import (
    get_config_path,
    get_ctfs_path,
    load_config,
    save_config,
)
from pwnv.cli.utils.crud import (
    add_challenge,
    add_ctf,
    add_tags,
    challenges_for_ctf,
    ctfs_with_challenges,
    get_challenges,
    get_ctf_by_challenge,
    get_ctfs,
    get_current_challenge,
    get_current_ctf,
    get_running_ctfs,
    get_solved_challenges,
    get_stopped_ctfs,
    get_tags,
    get_unsolved_challenges,
    is_duplicate,
    remove_challenge,
    remove_ctf,
    update_challenge,
    update_ctf,
)
from pwnv.cli.utils.guards import challenges_exists, config_exists, ctfs_exists
from pwnv.cli.utils.remote import (
    add_remote_ctf,
    normalise_category,
    remote_solve,
    sanitize,
)
from pwnv.cli.utils.ui import (
    command,
    error,
    info,
    prompt_category_selection,
    prompt_challenge_selection,
    prompt_confirm,
    prompt_ctf_selection,
    prompt_fuzzy_select,
    prompt_tags_selection,
    prompt_text,
    show_challenge,
    show_ctf,
    success,
    warn,
)

__all__ = [
    # config
    "get_config_path",
    "get_ctfs_path",
    "load_config",
    "save_config",
    # crud
    "add_challenge",
    "add_ctf",
    "add_tags",
    "challenges_for_ctf",
    "ctfs_with_challenges",
    "get_challenges",
    "get_ctf_by_challenge",
    "get_ctfs",
    "get_current_challenge",
    "get_current_ctf",
    "get_running_ctfs",
    "get_solved_challenges",
    "get_stopped_ctfs",
    "get_tags",
    "get_unsolved_challenges",
    "is_duplicate",
    "remove_challenge",
    "remove_ctf",
    "update_challenge",
    "update_ctf",
    # guards
    "challenges_exists",
    "config_exists",
    "ctfs_exists",
    # remote
    "add_remote_ctf",
    "normalise_category",
    "remote_solve",
    "sanitize",
    # ui
    "prompt_confirm",
    "error",
    "prompt_fuzzy_select",
    "prompt_category_selection",
    "prompt_challenge_selection",
    "prompt_ctf_selection",
    "prompt_tags_selection",
    "prompt_text",
    "show_challenge",
    "show_ctf",
    "success",
    "warn",
    "info",
    "command",
]



from typing import List, Sequence

from rich import print
from rich.markup import escape

from pwnv.cli.utils.crud import challenges_for_ctf, get_ctfs, get_tags
from pwnv.models import CTF, Challenge
from pwnv.models.challenge import Category, Solved


def success(msg: str):
    print(f"[green]✓[/] {msg}")


def error(msg: str):
    print(f"[red]error:[/] {msg}")


def warn(msg: str):
    print(f"[yellow]warning:[/] {msg}")


def info(msg: str):
    print(f"[blue]info:[/] {msg}")


def command(msg: str):
    return f"[cyan]`{msg}`[/]"


def _get_challenge_choices(challenges: Sequence[Challenge]):
    from InquirerPy.base.control import Choice

    ctf_names = {ctf.id: ctf.name for ctf in get_ctfs()}
    return [
        Choice(
            name=f"{ch.name:<50} [{ctf_names[ch.ctf_id]}]["
            f"{'solved' if ch.solved == Solved.solved else 'unsolved'}]"
            f"[{ch.category.name}]",
            value=ch,
        )
        for ch in challenges
    ]


def _get_ctf_choices(ctfs: Sequence[CTF]):
    from InquirerPy.base.control import Choice

    return [
        Choice(name=f"{ctf.name:<50} [{ctf.created_at.year}]", value=ctf)
        for ctf in ctfs
    ]


def prompt_confirm(message: str, default: bool = True, **kwargs):
    from InquirerPy import inquirer

    return inquirer.confirm(message=message, default=default, **kwargs).execute()


def prompt_fuzzy_select(
    *,
    choices,
    message: str = "Select:",
    **kwargs,
):
    from InquirerPy import inquirer

    return inquirer.fuzzy(
        message=message, choices=list(choices), border=True, **kwargs
    ).execute()


def prompt_challenge_selection(challenges: Sequence[Challenge], msg: str) -> Challenge:
    return prompt_fuzzy_select(
        choices=_get_challenge_choices(challenges),
        message=msg,
        transformer=lambda r: r.split(" ")[0],
    )


def prompt_ctf_selection(ctfs: Sequence[CTF], msg: str) -> CTF:
    return prompt_fuzzy_select(
        choices=_get_ctf_choices(ctfs),
        message=msg,
        transformer=lambda r: r.split(" ")[0],
    )


def prompt_category_selection() -> Category:
    category = prompt_fuzzy_select(
        choices=[c.name for c in Category], message="Select category:"
    )
    return Category[category]


def prompt_tags_selection(msg: str) -> List[str]:
    return prompt_fuzzy_select(choices=list(get_tags()), message=msg, multiselect=True)


def prompt_text(msg: str) -> str:
    from InquirerPy import inquirer

    return inquirer.text(message=msg).execute().strip()


def show_challenge(challenge: Challenge):
    print(f"[blue]{escape('[' + challenge.name + ']')}[/]")
    ctf = next(ctf for ctf in get_ctfs() if ctf.id == challenge.ctf_id)
    print(f"[red]ctf[/] = '{ctf.name}'")
    print(f"[red]category[/] = '{challenge.category.name}'")
    print(f"[red]path[/] = '{str(challenge.path)}'")
    print(f"[red]solved[/] = '{str(challenge.solved.name)}'")
    print(f"[red]points[/] = '{str(challenge.points)}'")
    print(f"[red]flag[/] = '{str(challenge.flag)}'")
    print(f"[red]tags[/] = '{', '.join(challenge.tags) if challenge.tags else ''}'")


def show_ctf(ctf: CTF):
    print(f"[blue]{escape('[' + ctf.name + ']')}[/]")
    print(f"[red]path[/] = '{str(ctf.path)}'")
    print(f"[red]running[/] = '{str(ctf.running.name)}'")
    print(f"[red]date[/] = '{str(ctf.created_at.date())}'")
    print(f"[red]num_challenges[/] = {len(challenges_for_ctf(ctf))}")



import shutil
import tempfile
from pathlib import Path

import typer
from rich import print

from pwnv.cli.utils import (
    command,
    config_exists,
    get_config_path,
    get_ctfs_path,
    info,
    prompt_confirm,
    success,
    warn,
)

app = typer.Typer(no_args_is_help=True)


@app.command()
@config_exists()
def reset() -> None:
    print("[red]" + "-" * 40 + " WARNING! " + "-" * 40 + "[/]")
    if not prompt_confirm(
        "This will delete the entire environment (config + files). Continue?",
        default=False,
    ):
        warn("Aborting reset.")
        return
    ctfs_path = get_ctfs_path()
    cfg_path = get_config_path()
    if prompt_confirm(
        "Do you want to backup the current environment as a tar.gz file?",
        default=False,
    ):
        backup_base = Path.home() / "pwnv_backup"
        backup_archive = backup_base.with_suffix(".tar.gz")

        if backup_archive.exists():
            if not prompt_confirm(
                f"Backup file already exists at {backup_archive}. Overwrite?",
                default=False,
            ):
                warn("Aborting backup creation.")
                return
            backup_archive.unlink()

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp = Path(tmpdir)

                def ignore_pwnvenv(dir, contents):
                    return [".pwnvenv"] if ".pwnvenv" in contents else []

                shutil.copytree(ctfs_path, tmp / ctfs_path.name, ignore=ignore_pwnvenv)
                shutil.copy(cfg_path, tmp / cfg_path.name)

                shutil.make_archive(str(backup_base), "gztar", root_dir=tmp)

            success(f"Backup created at {backup_archive}")
        except Exception as e:
            warn(f"Backup failed: {e}")

    if ctfs_path.exists():
        shutil.rmtree(ctfs_path)
        success(f"Removed CTF files at {ctfs_path}")

    else:
        info("No CTF files found - nothing to remove.")

    if cfg_path.exists():
        cfg_path.unlink()
        success(f"Removed config file at {cfg_path}")

    else:
        info("No config file found - nothing to remove.")

    success("Workspace reset complete!")

    info(f"Run {command('pwnv init')} to bootstrap a fresh environment.")



from pwnv.models import Challenge
from pwnv.plugins.plugin import ChallengePlugin
from pwnv.templates import pwn_template


class PwnPlugin(ChallengePlugin):
    category = "pwn"

    def create_template(self, challenge: Challenge):
        path = challenge.path
        with open(path / "solve.py", "w") as f:
            f.write(pwn_template)

    def logic(self, challenge: Challenge):
        return super().logic(challenge)



from abc import ABC, abstractmethod

from pwnv.models import Challenge


class ChallengePlugin(ABC):
    category: str

    def __init__(self):
        if not hasattr(self, "category") or not isinstance(self.category, str):
            raise TypeError("Plugin must define a `category` string.")

    @abstractmethod
    def create_template(self, challenge: Challenge) -> None:
        pass

    @abstractmethod
    def logic(self, challenge: Challenge) -> None:
        pass



from pathlib import Path

from pydantic import BaseModel

from pwnv.models.challenge import Challenge
from pwnv.models.ctf import CTF


class Init(BaseModel):
    ctfs_path: Path
    challenge_tags: list[str]
    ctfs: list[CTF] | list[None]
    challenges: list[Challenge] | list[None]



import json
import uuid
from datetime import datetime
from enum import IntEnum
from pathlib import Path

from pydantic import BaseModel


class Category(IntEnum):
    pwn = 1
    web = 2
    rev = 3
    crypto = 4
    steg = 5
    misc = 6
    osint = 7
    forensics = 8
    hardware = 9
    mobile = 10
    game = 11
    blockchain = 12
    other = 13


class Solved(IntEnum):
    unsolved = 0
    solved = 1


class Challenge(BaseModel):
    id: uuid.UUID | int = uuid.uuid4()
    name: str
    flag: str | None = None
    points: int | None = None
    solved: Solved = Solved.unsolved
    category: Category = Category.pwn
    ctf_id: uuid.UUID
    path: Path
    tags: list[str] | None = None
    extras: dict | None = None


class Status(IntEnum):
    running = 1
    stopped = 0


class CTF(BaseModel):
    id: uuid.UUID = uuid.uuid4()
    name: str
    created_at: datetime = datetime.now()
    path: Path
    running: Status = Status.running
    url: str | None = None
    username: str | None = None
    password: str | None = None
    token: str | None = None


class Init(BaseModel):
    ctfs_path: Path
    challenge_tags: list[str]
    ctfs: list[CTF] | list[None]
    challenges: list[Challenge] | list[None]


def migrate_config(old_config_path: Path, new_config_path: Path):
    with old_config_path.open() as f:
        data = json.load(f)

    if data.get("default_ctf_path", False):
        ctfs_path = Path.home() / "CTF"
    else:
        ctfs_path = Path(data.get("env_path", "."))

    challenge_tags = data.get("challenge_tags", [])

    ctfs = []
    for c in data.get("ctfs", []):
        ctfs.append(
            CTF(
                id=uuid.UUID(c["id"]),
                name=c["name"],
                created_at=datetime.fromisoformat(c["created_at"]),
                path=Path(c["path"]),
                running=Status(c["running"]),
                url=c.get("url"),
                username=c.get("username"),
                password=c.get("password"),
                token=c.get("token"),
            )
        )

    challenges = []
    for ch in data.get("challenges", []):
        challenges.append(
            Challenge(
                id=uuid.UUID(ch["id"]),
                name=ch["name"],
                flag=ch.get("flag"),
                points=ch.get("points"),
                solved=Solved(ch.get("solved", Solved.unsolved)),
                category=Category(ch.get("category", Category.other)),
                ctf_id=uuid.UUID(ch["ctf_id"]),
                path=Path(ch["path"]),
                tags=ch.get("tags"),
                extras=ch.get("extras"),
            )
        )

    init_model = Init(
        ctfs_path=ctfs_path,
        challenge_tags=challenge_tags,
        ctfs=ctfs,
        challenges=challenges,
    )

    with new_config_path.open("w", encoding="utf-8") as f:
        f.write(init_model.model_dump_json(indent=4))


if __name__ == "__main__":
    migrate_config(Path("config.json"), Path("new_config.json"))
    print("Migration complete. New config written to new_config.json")



from pwnv.cli.challenge import app as challenge_app
from pwnv.cli.ctf import app as ctf_app
from pwnv.cli.init import app as init_app
from pwnv.cli.reset import app as reset_app
from pwnv.cli.solve import app as solve_app

__all__ = ["challenge_app", "ctf_app", "init_app", "reset_app", "solve_app"]



import shutil
from pathlib import Path
from typing import List

from pwnv.cli.utils.config import load_config, save_config
from pwnv.models import CTF, Challenge
from pwnv.models.challenge import Solved


# -- [CRUD] --
# [READ]
def get_ctfs() -> List[CTF]:
    return [CTF(**ctf) for ctf in load_config().get("ctfs", [])]


def get_challenges() -> List[Challenge]:
    return [Challenge(**ch) for ch in load_config().get("challenges", [])]


def get_running_ctfs() -> List[CTF]:
    return [ctf for ctf in get_ctfs() if ctf.running]


def get_stopped_ctfs() -> List[CTF]:
    return [ctf for ctf in get_ctfs() if not ctf.running]


def get_unsolved_challenges() -> List[Challenge]:
    return [ch for ch in get_challenges() if ch.solved == Solved.unsolved]


def get_solved_challenges() -> List[Challenge]:
    return [ch for ch in get_challenges() if ch.solved == Solved.solved]


def challenges_for_ctf(ctf: CTF) -> List[Challenge]:
    return [ch for ch in get_challenges() if ch.ctf_id == ctf.id]


def ctfs_with_challenges() -> List[CTF]:
    populated = {ch.ctf_id for ch in get_challenges()}
    return [ctf for ctf in get_ctfs() if ctf.id in populated]


def get_ctf_by_challenge(ch: Challenge) -> CTF | None:
    for ctf in get_ctfs():
        if ch.ctf_id == ctf.id:
            return ctf
    return None


def get_tags() -> set[str]:
    return set(load_config().get("challenge_tags", []))


def get_current_ctf(path: Path = Path.cwd()) -> CTF | None:
    for ctf in get_ctfs():
        if path.is_relative_to(ctf.path):
            return ctf
    return None


def get_current_challenge(path: Path = Path.cwd()) -> Challenge | None:
    for ch in get_challenges():
        if path.is_relative_to(ch.path):
            return ch
    return None


# [CREATE]
def add_ctf(ctf: CTF) -> None:
    cfg = load_config()
    cfg.setdefault("ctfs", []).append(ctf.model_dump())
    save_config(cfg)
    ctf.path.mkdir(parents=True, exist_ok=True)


def add_challenge(ch: Challenge) -> None:
    from pwnv.setup import Core

    cfg = load_config()
    cfg.setdefault("challenges", []).append(ch.model_dump())
    save_config(cfg)
    ch.path.mkdir(parents=True, exist_ok=True)
    Core(ch)


def add_tags(tags: set[str]) -> None:
    cfg = load_config()
    cfg["challenge_tags"] = list(
        set(cfg.get("challenge_tags", [])) | {t.lower() for t in tags}
    )
    save_config(cfg)


# [UPDATE]
def update_ctf(ctf: CTF) -> None:
    cfg = load_config()
    ctfs = cfg.setdefault("ctfs", [])
    for idx, c in enumerate(ctfs):
        if c["id"] == str(ctf.id):
            ctfs[idx] = ctf.model_dump()
            break
    save_config(cfg)


def update_challenge(ch: Challenge) -> None:
    cfg = load_config()
    challenges = cfg.setdefault("challenges", [])
    for idx, item in enumerate(challenges):
        if item["id"] == str(ch.id):
            challenges[idx] = ch.model_dump()
            break
    save_config(cfg)


# [DELETE]
def remove_ctf(ctf: CTF) -> None:
    cfg = load_config()
    cfg["ctfs"] = [c for c in cfg.get("ctfs", []) if c["id"] != str(ctf.id)]
    cfg["challenges"] = [
        ch for ch in cfg.get("challenges", []) if ch["ctf_id"] != str(ctf.id)
    ]
    save_config(cfg)
    if ctf.path.exists():
        shutil.rmtree(ctf.path)


def remove_challenge(ch: Challenge) -> None:
    cfg = load_config()
    cfg["challenges"] = [
        item for item in cfg.get("challenges", []) if item["id"] != str(ch.id)
    ]
    save_config(cfg)
    if ch.path.exists():
        shutil.rmtree(ch.path)


# [MISC]
def is_duplicate(
    *,
    name: str | None = None,
    path: Path | None = None,
    model_list: List[CTF | Challenge],
) -> bool:
    if path is None:
        return any(model.name == name for model in model_list)
    elif name is None:
        return any(model.path == path for model in model_list)
    return any(model.name == name or model.path == path for model in model_list)



from functools import wraps


def _guard(predicate, msg):
    from pwnv.cli.utils.ui import warn

    def deco(fn):
        @wraps(fn)
        def wrapper(*a, **kw):
            if predicate():
                return fn(*a, **kw)
            warn(msg)

        return wrapper

    return deco


def config_exists():
    from pwnv.cli.utils.config import get_config_path
    from pwnv.cli.utils.ui import command

    return _guard(
        lambda: get_config_path().exists(),
        f"No config. Run {command('pwnv init')}. ",
    )


def ctfs_exists():
    from pwnv.cli.utils.crud import get_ctfs

    return _guard(lambda: bool(get_ctfs()), "No CTFs found.")


def challenges_exists():
    from pwnv.cli.utils.crud import get_challenges

    return _guard(lambda: bool(get_challenges()), "No challenges found.")



template = r"""#!/usr/bin/env python3

from pwn import *

con = ""

try:
    host, port = con.replace(" ", ":").split(":")
except:
    pass

ssl = False

binary = './binary'

gdbscript = '''
    c
'''

elf  = context.binary = ELF(binary)
libc = context.binary.libc
context.terminal = ['tmux', 'splitw', '-h']



# utils
u64 = lambda d: struct.unpack("<Q", d.ljust(8, b"\0"))[0]
u32 = lambda d: struct.unpack("<I", d.ljust(4, b"\0"))[0]
u16 = lambda d: struct.unpack("<H", d.ljust(2, b"\0"))[0]

# credits to spwn by @chino
ru         = lambda *x, **y: p.recvuntil(*x, **y, drop=True)
rl         = lambda *x, **y: p.recvline(*x, **y, keepends=False)
rc         = lambda *x, **y: p.recv(*x, **y)
sla        = lambda *x, **y: p.sendlineafter(*x, **y)
sa         = lambda *x, **y: p.sendafter(*x, **y)
sl         = lambda *x, **y: p.sendline(*x, **y)
sn         = lambda *x, **y: p.send(*x, **y)
logbase    = lambda: log.info("libc base = %#x" % libc.address)
logleak    = lambda name, val: log.info(name+" = %#x" % val)
ls         = lambda x: log.success(x)
lss        = lambda x: ls('\033[1;31;40m%s -> 0x%x \033[0m' % (x, eval(x)))
one_gadget = lambda: [int(i) + libc.address for i in subprocess.check_output(['one_gadget', '--raw', '-l1', libc.path]).decode().split(' ')]

# exit_handler stuff
fs_decrypt = lambda addr, key: ror(addr, 0x11) ^ key
fs_encrypt = lambda addr, key: rol(addr ^ key, 0x11)


# heap stuff
prot_ptr = lambda pos, ptr: (pos >> 12) ^ ptr
def deobfuscate(val):
    mask = 0xfff << 52
    while mask:
        v = val & mask
        val ^= (v >> 12)
        mask >>= 12
    return val


def start(argv=[], *a, **kw):
    if args.GDB: return gdb.debug([elf.path] + argv, gdbscript=gdbscript, *a, **kw)
    elif args.REMOTE: return remote(host, port, ssl=ssl)
    else: return process([elf.path] + argv, *a, **kw)


p = start()



p.interactive()


"""



import uuid
from datetime import datetime
from enum import IntEnum
from pathlib import Path

from pydantic import BaseModel


class Status(IntEnum):
    running = 1
    stopped = 0


class CTF(BaseModel):
    id: uuid.UUID = uuid.uuid4()
    name: str
    created_at: datetime = datetime.now()
    path: Path
    running: Status = Status.running
    url: str | None = None



import json
import os
from functools import lru_cache
from pathlib import Path
from tempfile import NamedTemporaryFile

from dotenv import load_dotenv
from filelock import SoftFileLock

load_dotenv()

_CONFIG_ENV: str = "PWNV_CONFIG"
_DEBUG_ENV: str = "PWNV_DEBUG"
_DEFAULT_BASENAME = "config.json"


def _resolve_config_path() -> Path:
    import typer

    if override := os.getenv(_CONFIG_ENV):
        return Path(override).expanduser().resolve()

    if os.getenv(_DEBUG_ENV):
        return Path("/tmp/pwnv") / _DEFAULT_BASENAME

    for parent in (Path.cwd(), *Path.cwd().parents):
        candidate = parent / _DEFAULT_BASENAME
        if candidate.is_file():
            return candidate

    return Path(typer.get_app_dir("pwnv")) / _DEFAULT_BASENAME


config_path: Path = _resolve_config_path()
config_path.parent.mkdir(parents=True, exist_ok=True)
_lock = SoftFileLock(str(config_path) + ".lock")


@lru_cache(maxsize=1)
def load_config() -> dict:
    if not config_path.exists():
        return {"ctfs": [], "challenges": [], "challenge_tags": []}
    with open(config_path) as f:
        return json.load(f)


def _invalidate_cache() -> None:
    load_config.cache_clear()


def save_config(cfg: dict) -> None:
    cfg.setdefault("ctfs", [])
    cfg.setdefault("challenges", [])
    cfg.setdefault("challenge_tags", [])

    with _lock:
        cfg_json = json.dumps(cfg, indent=4, default=str)
        with NamedTemporaryFile(
            "w", dir=config_path.parent, delete=False, encoding="utf-8"
        ) as tmp:
            tmp.write(cfg_json)
            tmp.flush()
            os.fsync(tmp.fileno())
        Path(tmp.name).replace(config_path)
    _invalidate_cache()


def get_config_path() -> Path:
    return config_path


def get_ctfs_path() -> Path:
    config = load_config()
    return Path(config["ctfs_path"])



import asyncio
import atexit
import os
import re

from dotenv import load_dotenv

from pwnv.cli.utils.ui import error, prompt_text, success, warn
from pwnv.models import CTF, Challenge
from pwnv.models.challenge import Category

_keyword_map = {
    "pwn": Category.pwn,
    "web": Category.web,
    "rev": Category.rev,
    "reverse": Category.rev,
    "crypto": Category.crypto,
    "cryptography": Category.crypto,
    "stego": Category.steg,
    "steganography": Category.steg,
    "misc": Category.misc,
    "miscellaneous": Category.misc,
    "osint": Category.osint,
    "forensics": Category.forensics,
    "hardware": Category.hardware,
    "mobile": Category.mobile,
    "game": Category.game,
    "blockchain": Category.blockchain,
}


def sanitize(name: str) -> str:
    return name.strip().replace(" ", "-").replace("..", ".").replace("/", "_").lower()


def normalise_category(raw: str) -> Category:
    clean = re.sub(r"\\(.*?\\)", "", raw).strip().lower()
    key = re.split(r"[^a-z]+", clean, maxsplit=1)[0]
    return _keyword_map.get(key, Category.other)


def _ask_for_credentials(methods) -> dict:
    from ctfbridge.models.auth import AuthMethod
    from InquirerPy import inquirer

    creds = {"username": None, "password": None, "token": None}
    chosen = inquirer.select(
        message="Choose authentication method:",
        choices=[method.name for method in methods],
    ).execute()
    if chosen == AuthMethod.CREDENTIALS.name:
        creds["username"] = prompt_text("Username:")
        creds["password"] = inquirer.secret(message="Password:").execute().strip()
    elif chosen == AuthMethod.TOKEN.name:
        creds["token"] = inquirer.secret(message="Token:").execute().strip()
    else:
        error("No supported authentication methods found.")
        return {}
    return creds


_runner: asyncio.Runner | None = None


def _run_async(coro):
    global _runner
    if _runner is None:
        _runner = asyncio.Runner()
        atexit.register(_runner.close)
    return _runner.run(coro)


def add_remote_ctf(ctf: CTF) -> None:
    from pwnv.cli.utils.crud import add_ctf, remove_ctf

    client, methods = _run_async(get_remote_credential_methods(ctf.url))
    if client is None:
        return
    creds = _ask_for_credentials(methods)
    if not creds:
        return

    add_ctf(ctf)

    if not _run_async(create_remote_session(client, creds, ctf)):
        remove_ctf(ctf)
        return

    challenges = _run_async(get_remote_challenges(client, ctf))
    if challenges is None:
        remove_ctf(ctf)
        return

    env_path = ctf.path / ".env"
    with open(env_path, "w") as f:
        if creds.get("username", None):
            f.write(f"CTF_USERNAME={creds.get('username')}\n")
        if creds.get("password", None):
            f.write(f"CTF_PASSWORD={creds.get('password')}\n")
        if creds.get("token", None):
            f.write(f"CTF_TOKEN={creds.get('token')}\n")

    _run_async(add_remote_challenges(client, ctf, challenges))


async def get_remote_credential_methods(url: str):
    from ctfbridge import create_client

    try:
        client = await create_client(url=url)
    except Exception:
        error("Failed to get client.")
        return None, None
    methods = await client.auth.get_supported_auth_methods()
    return client, methods


async def create_remote_session(client, creds, ctf) -> bool:
    try:
        await client.auth.login(**creds)
        await client.session.save(str(ctf.path / ".session"))
        return True
    except Exception:
        error("Failed to authenticate with the provided credentials.")
        return False


async def get_remote_challenges(client, ctf):
    try:
        await client.session.load(ctf.path / ".session")
        challenges = await client.challenges.get_all()
        return challenges
    except Exception:
        error("Failed to fetch challenges.")
        return None


async def add_remote_challenges(client, ctf: CTF, challenges) -> None:
    from pwnv.cli.utils.crud import add_challenge
    from pwnv.models.challenge import Solved

    for ch in challenges:
        category = normalise_category(ch.category)
        name = sanitize(ch.name)
        challenge = Challenge(
            name=name,
            ctf_id=ctf.id,
            path=ctf.path / category.name / name,
            category=category,
            points=ch.value,
            solved=Solved.solved if ch.solved else Solved.unsolved,
            extras={
                "slug": ch.id,
                "description": ch.description,
                "attachments": [att.model_dump() for att in ch.attachments],
                "author": ch.author,
            },
            tags=ch.tags,
        )
        add_challenge(challenge)
        try:
            await client.attachments.download_all(ch.attachments, challenge.path)
        except Exception:
            warn(f"Skipped attachments for {challenge.name}")

        success(f"{challenge.name} ({challenge.points} pts) added")


async def remote_solve(ctf: CTF, challenge: Challenge, flag: str) -> None:
    from ctfbridge import create_client

    client = await create_client(ctf.url)
    if (ctf.path / ".session").exists():
        try:
            await client.session.load(ctf.path / ".session")
        except Exception as e:
            warn(f"Ignoring broken session cookie ({e}).")

    elif (ctf.path / ".env").exists():
        load_dotenv(ctf.path / ".env")
        creds = {
            "username": os.getenv("CTF_USERNAME"),
            "password": os.getenv("CTF_PASSWORD"),
            "token": os.getenv("CTF_TOKEN"),
        }
        await client.auth.login(**creds)
    else:
        creds = _ask_for_credentials(await client.auth.get_supported_auth_methods())
        if not await create_remote_session(client, creds):
            return False

    try:
        res = await client.challenges.submit(challenge.extras["slug"], flag)
        if res.correct:
            success(f"Flag [cyan]{flag}[/] accepted!")

            return True
        else:
            error(f"Flag [cyan]{flag}[/] incorrect")
            return False
    except Exception:
        error(f"Failed to submit flag '{flag}'.")
        return False
